'''Time value of money functions'''


def future_value_series(
    payment: float,
    interest_rate: float,
    periods: int,
    compounding_frequency: int = 1,
    start_of_period: bool = False,
) -> float:
    '''
    Returns the total value of an future value series earning compound interest with regular additions.

        Parameters:
            payment (float): Value of the regular additions
            interest_rate (float): Interest rate per period, e.g. year
            periods (int): Number of additions and term of the investment, e.g. years
            compounding_frequency (int): Number of compoundings that occur per period, default 1
            start_of_period (bool): Make the payment at the start of each period, default False

        Returns:
            future_value (float): Value of the investment after the term
    '''
    if interest_rate == 0:
        return payment * periods

    effective_rate = interest_rate / compounding_frequency
    total_periods = periods * compounding_frequency
    start_modifier = 1
    if start_of_period:
        start_modifier = 1 + effective_rate
    return payment * start_modifier * (((1 + effective_rate) ** total_periods) - 1) / effective_rate


def present_value(
    payment: float,
    interest_rate: float,
    periods: int,
    future_value: float = 0,
    start_of_period: bool = False,
) -> float:
    '''
    Returns the present value of a loan or investment based on a constant interest rate.

        Parameters:
            payment (float): The regular payment made each period
            interest_rate (float): The interest rate per period, e.g. year
            periods (int): Number of payments made over the term of the investment
            future_value (float): The total cash amount you want to have at the last payment, default 0
            start_of_period (bool): Payment is made at start of each period, default False

        Returns:
            present_value (float): Present value of a loan or investment
    '''
    if interest_rate == 0:
        return (-1 * payment * periods) - future_value

    if start_of_period:
        pv_type = 1
    else:
        pv_type = 0

    numerator = payment * (1 + interest_rate * pv_type) * ((1 + interest_rate) ** periods - 1) / interest_rate + future_value
    denominator = (1 + interest_rate) ** periods

    return -1 * numerator / denominator


def discounted_cash_flow(cash_flows: list[float], discount_rate: float) -> float:
    '''
    Returns the discounted cash flow of a series of future cash flows.

        Parameters:
            cash_flows (list[float]): Future cash flows ordered chronologically
            discount_rate (float): Discount rate of the cash flows

        Returns:
            discounted_cash_flow (float): Adjusted present value of the future cash flows
    '''
    dcf = 0
    for i, cf in enumerate(cash_flows):
        dcf += cf / (1 + discount_rate) ** (i + 1)
    return dcf


def modified_internal_rate_of_return(cash_flows: list[float], finance_rate: float, reinvest_rate: float) -> float:
    '''
    Returns the modified internal rate of return for a list of periodic cash flows.
    Considers both the cost of investment and the interest received on reinvested cash.

        Parameters:
            cash_flows (list[float]): List of cash flows ordered chronologically. Must contain at least one positive
                                      and one negative value
            finance_rate (float): The interest rate you pay for money that is borrowed
            reinvest_rate (float): The interest rate you receive for money that is invested

        Returns:
            modified_internal_rate_of_return (float): Decimal value of the MIRR, None for invalid cash flows
    '''
    negative_values = []
    positive_values = []

    for value in cash_flows:
        if value < 0:
            negative_values.append(value)
            positive_values.append(0.0)
        elif value > 0:
            positive_values.append(value)
            negative_values.append(0.0)
        else:
            negative_values.append(0.0)
            positive_values.append(0.0)

    if (len(set(negative_values)) <= 1 or len(set(positive_values)) <= 1):
        return None

    n = len(cash_flows)
    numerator = -1 * discounted_cash_flow(positive_values, reinvest_rate) * (1 + reinvest_rate) ** n
    denominator = discounted_cash_flow(negative_values, finance_rate) * (1 + finance_rate)

    return (numerator / denominator) ** (1 / (n - 1)) - 1


def future_value_schedule(principal: float, interest_schedule: list[float]) -> float:
    '''
    Returns the future value of an investment based on a schedule of interest rates.

        Parameters:
            principal (float): The intiial investment sum
            interest_schedule (list[float]): Schedule of interest rates per period

        Returns:
            future_value (float): The future value of the investment
    '''
    for interest in interest_schedule:
        principal *= 1 + interest

    return principal



'''Unit and notation conversion functions'''


def dollar_decimal(fractional_dollar: float, fraction: int) -> float:
    '''
    Converts a fractional dollar into a decimal dollar.
    For example, a value of 1.3 with fraction 4 represents 1 + 3/4 = 1.75.

        Parameters:
            fractional_dollar (float): A number expressed as an integer portion and a fractional
                                       portion, separated by a decimal
            fraction (int): The denominator of the fractional portion, must be positive

        Returns:
            decimal_dollar (float): The dollar decimal representation of the fractional dollar
    '''
    fraction_length = len(str(fraction))
    integer_part = int(fractional_dollar)
    mantissa_part = (fractional_dollar - integer_part) * 10 ** fraction_length

    return(integer_part + mantissa_part / fraction)


def dollar_fractional(decimal_dollar: float, fraction: int) -> float:
    '''
    Converts a decimal dollar into a fractional dollar.
    For example, a value of 1.125 with fraction 16 repesents 1 + 12.5/100 = 1 + 2/16 = 1.02.

        Parameters:
            decimal_dollar (float): The decimal representation of the number
            fraction (int): The denominator of the fractional portion, must be positive

        Returns:
            fractional_dollar (float): The dollar fractional representation of the decimal dollar
    '''
    fraction_length = len(str(fraction))
    integer_part = int(decimal_dollar)
    mantissa_part = (decimal_dollar - integer_part) / 10 ** fraction_length

    return(integer_part + mantissa_part * fraction)



'''Common finance functions'''


def simple_interest(principle: float, interest_rate: float, periods: int) -> float:
    '''
    Returns the total value of an investment earning simple interest.

        Parameters:
            principle (float): Present value of the investment
            interest_rate (float): Interest rate per period, e.g. year
            periods (int): Term of the investment, e.g. years

        Returns:
            future_value (float): Value of the investment after the term
    '''
    interest = interest_rate * periods
    return principle * (1 + interest)


def compound_interest(principle: float, interest_rate: float, periods: int, compounding_frequency: int = 1) -> float:
    '''
    Returns the total value of an investment earning compound interest.

        Parameters:
            principle (float): Present value of the investment
            interest_rate (float): Interest rate per period, e.g. year
            periods (int): Term of the investment, e.g. years
            compounding_frequency (int): Number of compoundings that occur per period, default 1

        Returns:
            future_value (float): Value of the investment after the term
    '''
    return principle * (1 + effective_interest(interest_rate, compounding_frequency)) ** periods


def effective_interest(nominal_rate: float, periods: int) -> float:
    '''
    Returns the effective annual interest rate.

        Parameters:
            nominal_rate (float): The nominal interest rate (i.e. APR)
            periods (int): The number of compounding periods per year

        Returns:
            effective_interest: The effective interest rate
    '''
    return (1 + nominal_rate / periods) ** periods - 1


def loan_payment(principal: float, interest_rate: float, payment_frequency: int, term: int, down_payment: float = 0) -> float:
    '''
    Returns the periodic payment required to repay a loan accruing compound interest.

        Parameters:
            principal (float): Initial value of the loan
            interest_rate (float): Interest rate per period, e.g. year
            payment_frequency (int): Number of payments and compoundings per period, e.g. year
            term (int): Term of the loan in number of payments
            down_payment (float): Amount paid towards the loan before interest, default 0

        Returns:
            periodic_payment (float): Period loan payment
    '''
    loan_amount = principal - down_payment
    if interest_rate == 0:
        return loan_amount / term

    effective_rate = interest_rate / payment_frequency
    return loan_amount * effective_rate * (1 + effective_rate) ** term / ((1 + effective_rate) ** term - 1)


def equivalent_interest_rate(present_value: float, future_value: float, periods: int) -> float:
    '''
    Returns the equivalent interest rate for the growth of an investment.

        Parameters:
            present_value (float): The present value of the investment
            future_value (float): The future value of the investment
            periods (int): The number of periods elapsed between the future and present value

        Returns:
            equivalent_interest_rate (float): The equivalent interest rate
    '''
    return (future_value / present_value) ** (1 / periods) - 1


def loan_payment_schedule(
    principal: float,
    interest_rate: float,
    payment_frequency: int,
    term: int,
    down_payment: float = 0,
) -> dict[str, list[float]]:
    '''
    Returns the payment schedule for a loan.

        Parameters:
            principal (float): Initial value of the loan
            interest_rate (float): Interest rate per period, e.g. year
            payment_frequency (int): Number of payments and compoundings per period, e.g. year
            term (int): Term of the loan in number of payments
            down_payment (float): Amount paid towards the loan before interest, default 0

        Returns:
            loan_payment_schedule (dict):
                principal_payment (list[float]): Portion of the loan payment used to pay the principal
                interest_payment (list[float]): Portion of the loan payment used to pay the interest
                remaining_balance (list[float]): Remaining loan balance after payment
    '''
    payment = loan_payment(principal, interest_rate, payment_frequency, term, down_payment)
    loan_amount = principal - down_payment
    principal_payment, interest_payment, remaining_balance = [], [], []

    for _ in range(term):
        interest_payment.append(loan_amount * interest_rate / payment_frequency)
        principal_payment.append(payment - interest_payment[-1])
        remaining_balance.append(loan_amount - principal_payment[-1])
        loan_amount -= principal_payment[-1]

    return {
        'principal_payment': principal_payment,
        'interest_payment': interest_payment,
        'remaining_balance': remaining_balance,
    }



from pfinance import conversion, depreciation, general, securities, time_value


# Helper functions
def _compare_list_float(list1, list2, rounding_precision):
    # Compares two lists of floats after rounding.
    if len(list1) != len(list2):
        print("Lists are different lengths.")
        print("list1 length:", len(list1))
        print("list2 length:", len(list2))
        return False

    for i in range(len(list1)):
        if round(list1[i], rounding_precision) != round(list2[i], rounding_precision):
            print("List index", str(i), "have different values.")
            print("list1[" + str(i) + "] =", round(list1[i], rounding_precision))
            print("list2[" + str(i) + "] =", round(list2[i], rounding_precision))
            return False

    return True


def test_compare_list_float():
    list1 = [1.12, 1.22, 1.32, 1.42]
    list2 = [1.12, 1.22, 1.32]
    list3 = [1.123, 2.123, 3.123, 4.123]
    list4 = [1.123, 2.123, 3.126, 4.123]
    list5 = [2.457, 9.528, 9.65, 4.182]
    list6 = [2.46, 9.53, 9.65, 4.18]
    assert not(_compare_list_float(list1, list2, 1))  # Fail different lengths
    assert not(_compare_list_float(list3, list4, 2))  # Fail different values
    assert _compare_list_float(list5, list6, 2)


# General
def test_simple_interest():
    assert general.simple_interest(100, 0, 10) == 100.00
    assert general.simple_interest(100, 0.10, 10) == 200.00


def test_compound_interest():
    assert general.compound_interest(100, 0, 10) == 100.00
    assert round(general.compound_interest(100, 0.10, 10, 12), 2) == 270.70


def test_effective_interest():
    assert general.effective_interest(0, 12) == 0
    assert round(general.effective_interest(0.05, 12), 6) == 0.051162
    assert round(general.effective_interest(0.0525, 4), 7) == 0.0535427
    assert round(general.effective_interest(1.25, 7), 6) == 2.158576


def test_loan_payment():
    assert general.loan_payment(1000, 0, 1, 10) == 100.00
    assert round(general.loan_payment(100000, 0.10, 12, 60), 2) == 2124.70
    assert round(general.loan_payment(150000, 0.10, 12, 60, 50000), 2) == 2124.70


def test_equivalent_interest_rate():
    assert round(general.equivalent_interest_rate(10000, 11000, 96), 7) == 0.0009933
    assert round(general.equivalent_interest_rate(1000, 10000, 5), 3) == 0.585
    assert round(general.equivalent_interest_rate(700, 300, 12), 4) == -0.0682


def test_loan_payment_schedule():
    principal_payment1 = [162.55, 164.17, 165.82, 167.47, 169.15, 170.84]
    interest_payment1 = [10.0, 8.37, 6.73, 5.07, 3.40, 1.71]
    remaining_balance1 = [837.45, 673.28, 507.46, 339.99, 170.84, 0.00]
    principal_payment2 = [74.02, 74.39, 74.77, 75.14, 75.51, 75.89, 76.27]
    interest_payment2 = [2.63, 2.26, 1.89, 1.51, 1.14, 0.76, 0.38]
    remaining_balance2 = [451.98, 377.58, 302.82, 227.68, 152.16, 76.27, 0.0]
    principal_payment3 = [100.0, 100.0, 100.0, 100.0, 100.0]
    interest_payment3 = [0.0, 0.0, 0.0, 0.0, 0.0]
    remaining_balance3 = [400.0, 300.0, 200.0, 100.0, 0.0]
    assert _compare_list_float(general.loan_payment_schedule(1000, 0.12, 12, 6)['principal_payment'], principal_payment1, 2)
    assert _compare_list_float(general.loan_payment_schedule(1000, 0.12, 12, 6)['interest_payment'], interest_payment1, 2)
    assert _compare_list_float(general.loan_payment_schedule(1000, 0.12, 12, 6)['remaining_balance'], remaining_balance1, 2)
    assert _compare_list_float(general.loan_payment_schedule(526, 0.06, 12, 7)['principal_payment'], principal_payment2, 2)
    assert _compare_list_float(general.loan_payment_schedule(526, 0.06, 12, 7)['interest_payment'], interest_payment2, 2)
    assert _compare_list_float(general.loan_payment_schedule(526, 0.06, 12, 7)['remaining_balance'], remaining_balance2, 2)
    assert _compare_list_float(general.loan_payment_schedule(500, 0, 12, 5)['principal_payment'], principal_payment3, 2)
    assert _compare_list_float(general.loan_payment_schedule(500, 0, 12, 5)['interest_payment'], interest_payment3, 2)
    assert _compare_list_float(general.loan_payment_schedule(500, 0, 12, 5)['remaining_balance'], remaining_balance3, 2)


# Time Value
def test_future_value_series():
    assert time_value.future_value_series(100, 0, 10) == 1000.00
    assert round(time_value.future_value_series(100, 0.05, 10, 12), 2) == 15528.23
    assert round(time_value.future_value_series(100, 0.05, 10, 12, True), 2) == 15592.93


def test_present_value():
    assert time_value.present_value(100, 0, 12, 0, False) == -1200.00
    assert round(time_value.present_value(500, 0.06, 48, 9000, False), 2) == -8374.00
    assert round(time_value.present_value(200, 0.07, 36, 1000, True), 2) == -2877.07


def test_discounted_cash_flow():
    assert time_value.discounted_cash_flow([], 0) == 0.00
    assert round(time_value.discounted_cash_flow([1000, 1000, 4000, 4000, 6000], 0.05), 2) == 13306.73


def test_modified_internal_rate_of_return():
    assert time_value.modified_internal_rate_of_return([], 0.1, 0.1) is None
    assert time_value.modified_internal_rate_of_return([10, 10], 0.1, 0.1) is None
    assert time_value.modified_internal_rate_of_return([-10, -10], 0.1, 0.1) is None
    assert round(time_value.modified_internal_rate_of_return([-120000, 39000, 30000, 21000, 37000], 0.1, 0.12), 3) == 0.063
    assert round(time_value.modified_internal_rate_of_return([24, -96, -52, 27, -17, 15, -2, 0, 0], 0.05, 0.07), 3) == -0.056


def test_future_value_schedule():
    assert round(time_value.future_value_schedule(1000, [0.02, 0.03, 0.04, 0.05]), 2) == 1147.26
    assert round(time_value.future_value_schedule(123.45, [0, 0, 0, 0, 0, 0, 0, 0, 0]), 2) == 123.45
    assert round(time_value.future_value_schedule(15973, [0.02, 0.09, -0.08, 0.2]), 2) == 19605.69


# Conversion
def test_dollar_decimal():
    assert round(conversion.dollar_decimal(1.2, 16), 2) == 2.25
    assert round(conversion.dollar_decimal(9000.4123, 200), 4) == 9002.0615
    assert round(conversion.dollar_decimal(703.238, 23), 5) == 704.03478


def test_dollar_fractional():
    assert (round(conversion.dollar_fractional(1.125, 16), 2)) == 1.02
    assert (round(conversion.dollar_fractional(1.125, 32), 2)) == 1.04
    assert (round(conversion.dollar_fractional(738.526, 29), 5)) == 738.15254


# Depreciation
def test_straight_line_depreciation():
    assert round(depreciation.straight_line_depreciation(2000, 500, 5)) == 300
    assert depreciation.straight_line_depreciation(30000, 7500, 10) == 2250


def test_sum_of_years_depreciation():
    asset_value1 = [1000.0, 673.33, 412.00, 216.00, 85.33, 20.00]
    depreciation1 = [0.0, 326.67, 261.33, 196.00, 130.67, 65.33]
    asset_value2 = [12345, 9339.00, 6762.43, 4615.29, 2897.57, 1609.29, 750.43, 321.00]
    depreciation2 = [0.0, 3006.00, 2576.57, 2147.14, 1717.71, 1288.29, 858.86, 429.43]
    asset_value3 = [100.0, 0.0]
    depreciation3 = [0.0, 100.0]
    assert _compare_list_float(depreciation.sum_of_years_depreciation(1000, 20, 5)['asset_value'], asset_value1, 2)
    assert _compare_list_float(depreciation.sum_of_years_depreciation(1000, 20, 5)['periodic_depreciation'], depreciation1, 2)
    assert _compare_list_float(depreciation.sum_of_years_depreciation(12345, 321, 7)['asset_value'], asset_value2, 2)
    assert _compare_list_float(
        depreciation.sum_of_years_depreciation(12345, 321, 7)['periodic_depreciation'],
        depreciation2,
        2
    )
    assert _compare_list_float(depreciation.sum_of_years_depreciation(100, 0, 1)['asset_value'], asset_value3, 2)
    assert _compare_list_float(depreciation.sum_of_years_depreciation(100, 0, 1)['periodic_depreciation'], depreciation3, 2)


def test_double_declining_balance_depreciation():
    asset_value1 = [10000.0, 6000.0, 3600.0, 2160.0, 2000.0, 2000.0]
    depreciation1 = [0.0, 4000.0, 2400.0, 1440.0, 160.0, 0.0]
    asset_value2 = [20000.0, 10000.0, 5000.0, 2500.0, 1250.00, 1000.0, 1000.0]
    depreciation2 = [0.0, 10000.0, 5000.0, 2500.0, 1250.0, 250.0, 0.0]
    asset_value3 = [100.0, 100.0, 100.0]
    depreciation3 = [0.0, 0.0, 0.0]
    asset_value4 = [100.0, 0.0]
    depreciation4 = [0.0, 100.0]
    assert _compare_list_float(
        depreciation.double_declining_balance_depreciation(10000, 2000, 5)['asset_value'],
        asset_value1,
        2
    )
    assert _compare_list_float(
        depreciation.double_declining_balance_depreciation(10000, 2000, 5)['periodic_depreciation'],
        depreciation1,
        2
    )
    assert _compare_list_float(
        depreciation.double_declining_balance_depreciation(20000, 1000, 6, 3)['asset_value'],
        asset_value2,
        2
    )
    assert _compare_list_float(
        depreciation.double_declining_balance_depreciation(20000, 1000, 6, 3)['periodic_depreciation'],
        depreciation2,
        2
    )
    assert _compare_list_float(
        depreciation.double_declining_balance_depreciation(100, 200, 2)['asset_value'],
        asset_value3,
        2
    )
    assert _compare_list_float(
        depreciation.double_declining_balance_depreciation(100, 200, 2)['periodic_depreciation'],
        depreciation3,
        2
    )
    assert _compare_list_float(
        depreciation.double_declining_balance_depreciation(100, 0, 1)['asset_value'],
        asset_value4,
        2
    )
    assert _compare_list_float(
        depreciation.double_declining_balance_depreciation(100, 0, 1)['periodic_depreciation'],
        depreciation4,
        2
    )


def test_units_of_production_depreciation():
    assert depreciation.units_of_production_depreciation(25000, 0, 100, 4) == 1000.0
    assert depreciation.units_of_production_depreciation(500000, 20000, 240000, 10000) == 20000.0


def test_declining_balance():
    asset_value1 = [1000.0, 681.0, 463.76, 315.82, 215.07, 146.47, 99.74]
    depreciation1 = [0.0, 319.0, 217.24, 147.94, 100.75, 68.61, 46.72]
    asset_value2 = [500.0, 448.12, 262.15, 153.36, 89.72, 52.48, 30.70, 21.15]
    depreciation2 = [0.0, 51.88, 185.97, 108.79, 63.64, 37.23, 21.78, 9.56]
    asset_value3 = [900.0, 639.08, 321.45, 161.69, 81.33, 40.91, 32.44]
    depreciation3 = [0.0, 260.93, 317.62, 159.76, 80.36, 40.42, 8.47]
    assert _compare_list_float(depreciation.declining_balance(1000, 100, 6)['asset_value'], asset_value1, 2)
    assert _compare_list_float(depreciation.declining_balance(1000, 100, 6)['periodic_depreciation'], depreciation1, 2)
    assert _compare_list_float(depreciation.declining_balance(500, 20, 6, 3)['asset_value'], asset_value2, 2)
    assert _compare_list_float(depreciation.declining_balance(500, 20, 6, 3)['periodic_depreciation'], depreciation2, 2)
    assert _compare_list_float(depreciation.declining_balance(900, 29, 5, 7)['asset_value'], asset_value3, 2)
    assert _compare_list_float(depreciation.declining_balance(900, 29, 5, 7)['periodic_depreciation'], depreciation3, 2)


# Securities
def test_bond_coupon_rate():
    assert securities.bond_coupon_rate(1000, 0) == 0.00
    assert securities.bond_coupon_rate(1000, 10) == 0.01
    assert securities.bond_coupon_rate(1000, 25, 5) == 0.125


def norberts_gambit():
    assert securities.norberts_gambit(0, 0, 0)['base_value'] == 0
    assert securities.norberts_gambit(10, 50, 45)['base_value'] == 450
    assert securities.norberts_gambit(10, 50, 45)['base_gain'] == -50
    assert securities.norberts_gambit(10, 10, 9, 1.1)['base_value'] == 99
    assert securities.norberts_gambit(10, 10, 9, 1.1)['base_gain'] == -1
    assert securities.norberts_gambit(10, 10, 9, 1.1)['converted_value'] == 90
    assert round(securities.norberts_gambit(10, 10, 9, 1.1)['converted_gain'], 2) == 90.91
    assert securities.norberts_gambit(20, 15, 10, 1.5, 7.5, 5)['base_value'] == 292.5
    assert securities.norberts_gambit(20, 15, 10, 1.5, 7.5, 5)['base_gain'] == -15
    assert securities.norberts_gambit(20, 15, 10, 1.5, 7.5, 5)['converted_value'] == 195
    assert securities.norberts_gambit(20, 15, 10, 1.5, 7.5, 5)['converted_gain'] == -10


def test_adjusted_cost_base():
    test_acb = securities.adjusted_cost_base()
    test_acb.buy(10, 10.00, 5.00)
    assert round(test_acb.get_acb(), 2) == 10.50
    assert round(test_acb.sell(5, 15.00, 5.00), 2) == 17.5
    assert round(test_acb.get_acb(), 2) == 10.50
    test_acb.buy(5, 20.00, 5.00)
    assert round(test_acb.get_acb(), 2) == 15.75
    assert round(test_acb.sell(10, 10.00, 5.00), 2) == -62.50
    assert round(test_acb.get_acb(), 2) == 0.00



'''Securities tracking and analysis functions'''


def bond_coupon_rate(face_value: float, payment: float, payment_rate: int = 1) -> float:
    '''
    Returns the coupon rate of a bond.

        Parameters:
            face_value (float): Par value of the bond at issue
            payment (float): Periodic payment from the bond
            payment_rate (int): Payments per interest period, default 1

        Returns:
            coupon_rate (float): Effective interest rate returned by the bond
    '''
    return (payment * payment_rate) / face_value


def norberts_gambit(
    quantity: int,
    purchase_price: float,
    sale_price: float,
    rate: float = 1,
    purchase_commission: float = 0,
    sale_commission: float = 0,
) -> dict[str, float]:
    '''
    Returns the converted value and capital gain of an execution of Norbert's Gambit.

        Parameters:
            quantity (int): Number of securities transacted
            purchase_price (float): Unit price of the purchased securities
            sale_price (float): Unit price of the sold securities
            rate (float): Exchange rate between the purchasing currency and the sale currency expressed as a multiplier of the
                          purchasing currency, default 1
            purchase_commission (float): Commission paid in the purchasing currency on the purchase transaction, default 0
            sale_commission (float): Commission paid in the sale currency on the sale transaction, default 0

        Returns:
            gambit_result (dict):
                base_value (float): Final value of the conversion expressed in the purchase currency
                base_gain (float): Capital gain of the conversion expressed in the purchase currency
                converted_value (float): Final value of the conversion expressed in the sale currency
                converted_gain (float): Capital gain of the conversion expressed in the sale currency
    '''
    initial_value = quantity * purchase_price - purchase_commission
    final_value = quantity * sale_price - sale_commission

    return {
        'base_value': final_value / rate,
        'base_gain': (final_value / rate) - initial_value,
        'converted_value': final_value,
        'converted_gain': final_value - (initial_value * rate),
    }


class adjusted_cost_base:
    '''
    Represents an adjusted cost base tracker

    Methods:
        buy(quantity, unit_price, commission): Records a purchase transaction
        sell(quantity, unit_price, commission): Records a sale transaction
        get_acb(): Returns the adjusted cost base of the position
    '''
    def __init__(self):
        '''
        Constructs the necessary attributes for the adjusted cost base tracker object.
        '''
        self._book_value = 0
        self._acb = 0
        self._shares = 0

    def buy(self, quantity: int, unit_price: float, commission: float = 0):
        '''
        Records a purchase transaction.

        Parameters:
            quantity (int): Number of securities purchased
            unit_price (float): Price per security paid
            commission (float): Commission paid in the transaction, default 0
        '''
        self._shares += quantity
        self._book_value += quantity * unit_price + commission
        self._acb = self._book_value / self._shares

    def sell(self, quantity: int, unit_price: float, commission: float = 0) -> float:
        '''
        Records a sale transaction.

        Parameters:
            quantity (int): Number of securities sold
            unit_price (float): Price per security received
            commission (float): Commission paid in the transaction, default 0

        Returns:
            capital_gain (float): The capital gain on the sale
        '''
        self._shares -= quantity
        capital_gain = (quantity * unit_price - commission) - (quantity * self._acb)
        if self._shares == 0:
            self._book_value = 0
            self._acb = 0
        else:
            self._book_value -= quantity * self._acb
        return capital_gain

    def get_acb(self) -> float:
        '''
        Returns the adjusted cost base of the position

        Returns:
            acb (float): Adjusted cost base of the position
        '''
        return self._acb



'''Depreciation of assets functions'''


def straight_line_depreciation(purchase_price: float, salvage_value: float, useful_life: int) -> float:
    '''
    Calculates the constant periodic depreciation of an asset.

        Parameters:
            purchase_price (float): The total amount paid for the asset
            salvage_value (float): The value of the asset after its useful life
            useful_life (int): The expected lifespan of an asset in periods, must be greater than 0

        Returns:
            periodic_depreciation (float): The periodic decrease in value of the asset
    '''
    return (purchase_price - salvage_value) / useful_life


def sum_of_years_depreciation(purchase_price: float, salvage_value: float, useful_life: int) -> dict[str, list[float]]:
    '''
    Calculates the depreciation of an asset using sum of years depreciation.

        Parameters:
            purchase_price (float): The total amount paid for the asset
            salvage_value (float): The value of the asset after its useful life
            useful_life (int): The expected lifespan of an asset, must be greater than 0

        Returns:
            sum_of_years_result (dict):
                asset_value (list[float]): Value of the asset at the beginning of the period
                periodic_depreciation (list[float]): Depreciation of the asset at the end of the period
    '''
    total_years = useful_life * (useful_life + 1) / 2
    asset_value = [float(purchase_price)]
    periodic_depreciation = [0.0]

    for i in range(useful_life, 0, -1):
        current_depreciation = (purchase_price - salvage_value) * i / total_years
        periodic_depreciation.append(current_depreciation)
        asset_value.append(asset_value[-1] - current_depreciation)

    return {
        'asset_value': asset_value,
        'periodic_depreciation': periodic_depreciation
    }


def double_declining_balance_depreciation(
    purchase_price: float,
    salvage_value: float,
    useful_life: int,
    factor: float = 2.0,
) -> dict[str, list[float]]:
    '''
    Calculates the depreciation of an asset using double declining balance.

        Parameters:
            purchase_price (float): The total amount paid for the asset
            salvage_value (float): The value of the asset after its useful life
            useful_life (int): The expected lifespan of an asset, must be greater than 0
            factor (int): The rate at which the balance declines, default 2

        Returns:
            double_declining_balance_result (dict):
                asset_value (list[float]): Value of the asset at the beginning of the period
                periodic_depreciation (list[float]): Depreciation of the asset at the end of the period
    '''
    asset_value = [float(purchase_price)]
    periodic_depreciation = [0.0]
    total_depreciation = 0.0

    for _ in range(useful_life):
        current_depreciation = max(
            0,
            min(
                (purchase_price - total_depreciation) * factor / useful_life,
                purchase_price - salvage_value - total_depreciation,
            )
        )

        periodic_depreciation.append(current_depreciation)
        total_depreciation += current_depreciation
        asset_value.append(asset_value[-1] - current_depreciation)

    return {
        'asset_value': asset_value,
        'periodic_depreciation': periodic_depreciation,
    }


def units_of_production_depreciation(
    purchase_price: float,
    salvage_value: float,
    useful_life: int,
    units_produced: int,
) -> float:
    '''
    Calculates the depreciation for a given period using units of production depreciation.

        Parameters:
            purchase_price (float): The total amount paid for the asset
            salvage_value (float): The value of the asset after its useful life
            useful_life (int): The total number of units the asset is expected to produce
            units_produced (int): The number of units produced by the asset during a single period

        Returns:
            depreciation (float): The depreciation of the asset for a single period
    '''
    return (units_produced / useful_life) * (purchase_price - salvage_value)


def declining_balance(
    purchase_price: float,
    salvage_value: float,
    useful_life: int,
    sub_periods: int = 12,
) -> dict[str, list[float]]:
    '''
    Calculates the depreciation of an asset using declining balance.

        Parameters:
            purchase_price (float): The total amount paid for the asset
            salvage_value (float): The value of the asset after its useful life
            useful_life (int): The expected lifespan of an asset, must be greater than 0
            sub_periods (int): The number of sub periods inside the first period (e.g. 12 months in one year). Must be less
                               than or equal to 12, default 12

        Returns:
            declining_balance_result (dict):
                asset_value (list[float]): Value of the asset at the beginning of the period
                periodic_depreciation (list[float]): Depreciation of the asset at the end of the period
    '''
    rate = round(1 - (salvage_value / purchase_price) ** (1 / useful_life), 3)  # Excel rounds rate to 3 decimal places
    asset_value = [float(purchase_price)]
    periodic_depreciation = [0.0]
    periodic_depreciation.append(purchase_price * rate * sub_periods / 12)  # First depreciation is speical case

    for _ in range(useful_life - 1):
        asset_value.append(asset_value[-1] - periodic_depreciation[-1])
        periodic_depreciation.append(asset_value[-1] * rate)

    asset_value.append(asset_value[-1] - periodic_depreciation[-1])

    if sub_periods != 12:  # Last depreciation is special case
        periodic_depreciation.append(asset_value[-1] * rate * (12 - sub_periods) / 12)
        asset_value.append(asset_value[-1] - periodic_depreciation[-1])

    return {
        'asset_value': asset_value,
        'periodic_depreciation': periodic_depreciation,
    }



import nox


@nox.session
def lint(session):
    targets = ['pfinance', 'tests', 'noxfile.py']
    session.install('flake8')
    session.run('flake8', *targets)


@nox.session
def tests(session):
    session.install('pytest')
    session.run('pytest', '-v')
